@@ -7,7 +7,7 @@ let token = Dune_lexer.token
 let jbuild_token = Jbuild_lexer.token
 
 let of_syntax = function
-  | Syntax.Dune -> token
+  | File_syntax.Dune -> token
   | Jbuild -> jbuild_token
 
 exception Error = Lexer_shared.Error
