@@ -32,34 +32,31 @@ let explode_path =
       | xs -> xs
 
 module External : sig
-  type t
-
-  val to_sexp : t Sexp.Encoder.t
-  val to_dyn : t -> Dyn.t
-
-  val compare : t -> t -> Ordering.t
+  include Path_intf.S
   val compare_val : t -> t -> Ordering.t
-  val to_string : t -> string
-  val of_string : string -> t
   val relative : t -> string -> t
   val mkdir_p : t -> unit
-  val basename : t -> string
-  val parent : t -> t
   val initial_cwd : t
   val cwd : unit -> t
-  val extend_basename : t -> suffix:string -> t
-  val extension : t -> string
-  val is_suffix : t -> suffix:string -> bool
-  val split_extension : t -> t * string
-  val set_extension : t -> ext:string -> t
   val as_local : t -> string
 end = struct
-  include Interned.No_interning(struct
+  module T = Interned.No_interning(struct
       let initial_size = 512
       let resize_policy = Interned.Greedy
       let order = Interned.Natural
     end)()
 
+  module Table = Hashtbl.Make(T)
+
+  type t = T.t
+
+  let to_string = T.to_string
+  let make = T.make
+  let equal = T.equal
+  let hash = T.hash
+  let pp = T.pp
+  let compare = T.compare
+
   let compare_val x y = String.compare (to_string x) (to_string y)
 
   let as_string x ~f =
@@ -77,6 +74,11 @@ end = struct
         [ "t", Sexp.Encoder.string t ];
     make t
 
+  let parse_string_exn ~loc t =
+    if Filename.is_relative t then
+      Exn.fatalf ~loc "path %s is not absolute" t;
+    make t
+
   let to_sexp t = Sexp.Encoder.string (to_string t)
   let to_dyn t = Dyn.String (to_string t)
 
@@ -118,7 +120,16 @@ end = struct
         Unix.mkdir t_s 0o777
 
   let basename t = Filename.basename (to_string t)
-  let parent t = as_string ~f:Filename.dirname t
+
+  let root = of_string "/"
+
+  let is_root = equal root
+
+  let parent t =
+    if is_root t then
+      None
+    else
+      Some (as_string t ~f:Filename.dirname)
 
   let extension t = Filename.extension (to_string t)
   let split_extension t =
@@ -136,41 +147,53 @@ end = struct
   let as_local t =
     let s = to_string t in
     "." ^ s
+
+  include (
+    Comparable.Operators(struct
+      type nonrec t = t
+      let compare = compare_val
+    end)
+    : Comparable.OPS with type t := t
+  )
+
+  let to_string_maybe_quoted t =
+    String.maybe_quoted (to_string t)
+
+  let parent_exn t =
+    match parent t with
+    | None -> Exn.code_error "Path.External.parent_exn called on a root path" []
+    | Some p -> p
+
+  let is_descendant b ~of_:a =
+    if is_root a then true
+    else
+      String.is_prefix ~prefix:(to_string a ^ "/") (to_string b)
+
+  module Set = struct
+    include T.Set
+    let of_listing ~dir ~filenames =
+      of_list (List.map filenames ~f:(fun f -> relative dir f))
+  end
+
+  module Map = T.Map
 end
 
-module Local : sig
-  type t
-
-  val to_sexp : t Sexp.Encoder.t
-  val to_dyn : t -> Dyn.t
+module Relative : sig
+  include Path_intf.S
 
   val root : t
   val is_root : t -> bool
-  val compare : t -> t -> Ordering.t
   val compare_val : t -> t -> Ordering.t
-  val equal : t -> t -> bool
-  val of_string : ?error_loc:Loc0.t -> string -> t
-  val to_string : t -> string
   val relative : ?error_loc:Loc0.t -> t -> string -> t
   val append : t -> t -> t
-  val parent : t -> t
   val mkdir_p : t -> unit
   val descendant : t -> of_:t -> t option
   val is_descendant : t -> of_:t -> bool
   val reach : t -> from:t -> string
-  val basename : t -> string
-  val extend_basename : t -> suffix:string -> t
-  val extension : t -> string
-  val is_suffix : t -> suffix:string -> bool
-  val split_extension : t -> t * string
-  val pp : Format.formatter -> t -> unit
-
-  val set_extension : t -> ext:string -> t
 
   module L : sig
     val relative : ?error_loc:Loc0.t -> t -> string list -> t
   end
-  module Set : Set.S with type elt = t
 
   module Prefix : sig
     type local = t
@@ -182,27 +205,36 @@ module Local : sig
     (* for all local path p, drop (invalid p = None) *)
     val invalid : t
   end with type local := t
+
+  val split_first_component : t -> (string * t) option
+  val explode : t -> string list
+  val of_relative : t -> t
+
 end = struct
-  (* either "." for root, either a '/' separated list of components
+  (* either "." for root, or a '/' separated list of components
      other that ".", ".."  and not containing '/'. *)
-  include Interned.No_interning(struct
+  module T = Interned.No_interning(struct
       let initial_size = 512
       let resize_policy = Interned.Greedy
       let order = Interned.Natural
     end)()
 
+  module Table = Hashtbl.Make(T)
+
+  type t = T.t
+
+  let to_string = T.to_string
+  let make = T.make
+  let hash = T.hash
+  let compare = T.compare
+
   let pp ppf s = Format.pp_print_string ppf (to_string s)
 
   let compare_val x y = String.compare (to_string x) (to_string y)
 
-  let equal x y =
-    match compare x y with
-    | Eq -> true
-    | Gt | Lt -> false
-
   let root = make "."
 
-  let is_root t = t = root
+  let is_root t = Ordering.is_eq (compare t root)
 
   let is_suffix t ~suffix = String.is_suffix (to_string t) ~suffix
 
@@ -225,12 +257,17 @@ end = struct
 
   let parent t =
     if is_root t then
-      Exn.code_error "Path.Local.parent called on the root" []
+      None
     else
       let t = to_string t in
       match String.rindex_from t (String.length t - 1) '/' with
-      | exception Not_found -> root
-      | i -> make (String.take t i)
+      | exception Not_found -> Some root
+      | i -> Some (make (String.take t i))
+
+  let parent_exn t =
+    match parent t with
+    | None -> Exn.code_error "Path.Local.parent called on the root" []
+    | Some t -> t
 
   let basename t =
     if is_root t then
@@ -255,7 +292,10 @@ end = struct
           if is_root t then
             Result.Error ()
           else
-            loop (parent t) rest
+            begin match parent t with
+            | None -> Error ()
+            | Some parent -> loop parent rest
+            end
         | fn :: rest ->
           if is_root t then
             loop (make fn) rest
@@ -322,12 +362,12 @@ end = struct
       let len = String.length s in
       len = 0 || before_slash s (len - 1)
 
-  let of_string ?error_loc s =
+  let parse_string_exn ~loc s =
     match s with
     | "" | "." -> root
     | _ when is_canonicalized s -> make s
-    | _ ->
-      relative root s ?error_loc
+    | _ -> relative root s ~error_loc:loc
+  let of_string s = parse_string_exn ~loc:Loc0.none s
 
   let rec mkdir_p t =
     if is_root t then
@@ -339,7 +379,7 @@ end = struct
       with
       | Unix.Unix_error (EEXIST, _, _) -> ()
       | Unix.Unix_error (ENOENT, _, _) as e ->
-        let parent = parent t in
+        let parent = parent_exn t in
         if is_root parent then
           raise e
         else begin
@@ -437,8 +477,55 @@ end = struct
       ; path_slash = "/"
       }
   end
+
+  include (
+    Comparable.Operators(struct
+      type nonrec t = t
+      let compare = compare_val
+    end)
+    : Comparable.OPS with type t := t
+  )
+
+  let split_first_component t =
+    if is_root t then None
+    else
+      let t = (to_string t) in
+      begin match String.lsplit2 t ~on:'/' with
+      | None -> Some (t, root)
+      | Some (before, after) ->
+        Some
+          ( before
+          , after
+            |> of_string)
+      end
+
+  let explode p =
+    if is_root p then []
+    else String.split (to_string p) ~on:'/'
+
+  let to_string_maybe_quoted t =
+    String.maybe_quoted (to_string t)
+
+  let parent_exn t =
+    match parent t with
+    | None -> Exn.code_error "Path.Relative.parent:exn t is root"
+                ["t", to_sexp t]
+    | Some parent -> parent
+
+  let of_relative t = t
+
+  module Set = struct
+    include T.Set
+    let of_listing ~dir ~filenames =
+      of_list (List.map filenames ~f:(fun f -> relative dir f))
+  end
+
+  module Map = T.Map
 end
 
+module Local = Relative
+module Source0 = Relative
+
 let (abs_root, set_root) =
   let root_dir = ref None in
   let set_root new_root =
@@ -458,6 +545,7 @@ let (abs_root, set_root) =
   in
   (abs_root, set_root)
 
+
 module Kind = struct
   type t =
     | External of External.t
@@ -497,64 +585,131 @@ module Kind = struct
     | Local t -> Local.mkdir_p t
     | External t -> External.mkdir_p t
 
-  let append_local x y =
+  let append_relative x y =
     match x with
     | Local x -> Local (Local.append x y)
     | External x -> External (External.relative x (Local.to_string y))
 
 end
 
-let (build_dir_kind, build_dir_prefix, set_build_dir) =
-  let build_dir = ref None in
-  let build_dir_prefix = ref None in
-  let set_build_dir (new_build_dir : Kind.t) =
-    match !build_dir with
+
+module Build = struct
+  include Relative
+  let append_source = append
+  let append_relative = append
+  let append_local = append
+
+  let local t = t
+
+  let extract_build_context t =
+    let t = Local.to_string t in
+    begin match String.lsplit2 t ~on:'/' with
     | None ->
-      (match new_build_dir with
-       | External _ -> ()
-       | Local p ->
-         if Local.is_root p || Local.parent p <> Local.root then
-           Exn.fatalf
-             "@{<error>Error@}: Invalid build directory: %s\n\
-              The build directory must be an absolute path or \
-              a sub-directory of the root of the workspace."
-             (Local.to_string p |> String.maybe_quoted));
-      build_dir := Some new_build_dir;
-      build_dir_prefix :=
-        Some (match new_build_dir with
-          | Local    p -> Local.Prefix.make p
-          | External _ -> Local.Prefix.invalid)
-    | Some build_dir ->
-      Exn.code_error "set_build_dir: cannot set build_dir more than once"
-        [ "build_dir", Kind.to_sexp build_dir
-        ; "new_build_dir", Kind.to_sexp new_build_dir ]
-  in
-  let build_dir = lazy (
-    match !build_dir with
+      Some (t, Source0.root)
+    | Some (before, after) ->
+      Some
+        ( before
+        , after
+          |> Source0.of_string )
+    end
+
+  let extract_build_context_dir t =
+    let t_str = Local.to_string t in
+    begin match String.lsplit2 t_str ~on:'/' with
+    | None -> Some (t, Source0.root)
+    | Some (before, after) ->
+      Some
+        ( Local.of_string before
+        , after
+          |> Source0.of_string
+        )
+    end
+
+  let extract_build_context_dir_exn t =
+    match extract_build_context_dir t with
+    | Some t -> t
+    | None -> Exn.code_error "Path.Build.extract_build_context_dir_exn"
+                ["t", to_sexp t]
+
+  let extract_build_context_exn t =
+    match extract_build_context t with
+    | Some t -> t
+    | None -> Exn.code_error "Path.Build.extract_build_context_exn"
+                ["t", to_sexp t]
+
+  let drop_build_context t = Option.map (extract_build_context t) ~f:snd
+  let drop_build_context_exn t =
+    match drop_build_context t with
+    | Some d -> d
     | None ->
-      Exn.code_error "build_dir: cannot use build dir before it's set" []
-    | Some build_dir -> build_dir)
-  in
-  let build_dir_prefix = lazy (
-    match !build_dir_prefix with
-    | None ->
-      Exn.code_error "build_dir: cannot use build dir before it's set" []
-    | Some prefix -> prefix)
-  in
-  (build_dir, build_dir_prefix, set_build_dir)
+      Exn.code_error "Path.Build.drop_build_context_exn"
+        [ "t", to_sexp t
+        ]
+
+  let is_alias_stamp_file s =
+    String.is_prefix (Local.to_string s) ~prefix:".aliases/"
+
+  let (build_dir_kind, build_dir_prefix, set_build_dir) =
+    let build_dir = ref None in
+    let build_dir_prefix = ref None in
+    let set_build_dir (new_build_dir : Kind.t) =
+      match !build_dir with
+      | None ->
+        (match new_build_dir with
+         | External _ -> ()
+         | Local p ->
+           if Local.is_root p || Local.parent_exn p <> Local.root then
+             Exn.fatalf
+               "@{<error>Error@}: Invalid build directory: %s\n\
+                The build directory must be an absolute path or \
+                a sub-directory of the root of the workspace."
+               (Local.to_string p |> String.maybe_quoted));
+        build_dir := Some new_build_dir;
+        build_dir_prefix :=
+          Some (match new_build_dir with
+            | Local    p -> Local.Prefix.make p
+            | External _ -> Local.Prefix.invalid)
+      | Some build_dir ->
+        Exn.code_error "set_build_dir: cannot set build_dir more than once"
+          [ "build_dir", Kind.to_sexp build_dir
+          ; "new_build_dir", Kind.to_sexp new_build_dir ]
+    in
+    let build_dir = lazy (
+      match !build_dir with
+      | None ->
+        Exn.code_error "build_dir: cannot use build dir before it's set" []
+      | Some build_dir -> build_dir)
+    in
+    let build_dir_prefix = lazy (
+      match !build_dir_prefix with
+      | None ->
+        Exn.code_error "build_dir: cannot use build dir before it's set" []
+      | Some prefix -> prefix)
+    in
+    (build_dir, build_dir_prefix, set_build_dir)
+
+  let to_string p =
+    match Lazy.force build_dir_kind with
+    | Local    b -> Local.to_string (Local.append b p)
+    | External b ->
+      if Local.is_root p then
+        External.to_string b
+      else
+        Filename.concat (External.to_string b) (Local.to_string p)
+end
 
 module T : sig
   type t = private
     | External of External.t
-    | In_source_tree of Local.t
-    | In_build_dir of Local.t
+    | In_source_tree of Relative.t
+    | In_build_dir of Relative.t
 
   val compare : t -> t -> Ordering.t
   val equal : t -> t -> bool
   val hash : t -> int
 
-  val in_build_dir : Local.t -> t
-  val in_source_tree : Local.t -> t
+  val in_build_dir : Relative.t -> t
+  val in_source_tree : Relative.t -> t
   val external_ : External.t -> t
 end = struct
   type t =
@@ -594,7 +749,7 @@ let is_root = function
 module Map = Map.Make(T)
 
 let kind = function
-  | In_build_dir p -> Kind.append_local (Lazy.force build_dir_kind) p
+  | In_build_dir p -> Kind.append_relative (Lazy.force Build.build_dir_kind) p
   | In_source_tree s -> Kind.Local s
   | External s -> Kind.External s
 
@@ -607,14 +762,7 @@ let to_string t =
   match t with
   | In_source_tree p -> Local.to_string p
   | External       p -> External.to_string p
-  | In_build_dir   p ->
-    match Lazy.force build_dir_kind with
-    | Local    b -> Local.to_string (Local.append b p)
-    | External b ->
-      if Local.is_root p then
-        External.to_string b
-      else
-        Filename.concat (External.to_string b) (Local.to_string p)
+  | In_build_dir   p -> Build.to_string p
 
 let to_string_maybe_quoted t =
   String.maybe_quoted (to_string t)
@@ -622,7 +770,7 @@ let to_string_maybe_quoted t =
 let root = in_source_tree Local.root
 
 let make_local_path p =
-  match Local.Prefix.drop (Lazy.force build_dir_prefix) p with
+  match Local.Prefix.drop (Lazy.force Build.build_dir_prefix) p with
   | None -> in_source_tree p
   | Some p -> in_build_dir p
 let of_local = make_local_path
@@ -639,14 +787,16 @@ let relative ?error_loc t fn =
     | In_build_dir p -> in_build_dir (Local.relative p fn ?error_loc)
     | External s -> external_ (External.relative s fn)
 
-let of_string ?error_loc s =
+let parse_string_exn ~loc s =
   match s with
   | "" | "." -> in_source_tree Local.root
   | s  ->
     if Filename.is_relative s then
-      make_local_path (Local.of_string s ?error_loc)
+      make_local_path (Local.parse_string_exn ~loc s)
     else
-      external_ (External.of_string s)
+      external_ (External.parse_string_exn ~loc s)
+
+let of_string s = parse_string_exn ~loc:Loc0.none s
 
 let to_sexp t =
   let constr f x y = Sexp.Encoder.(pair string f) (x, y) in
@@ -684,18 +834,18 @@ let reach t ~from =
   | In_source_tree t, In_source_tree from
   | In_build_dir   t, In_build_dir   from -> Local.reach t ~from
   | In_source_tree t, In_build_dir from -> begin
-      match Lazy.force build_dir_kind with
+      match Lazy.force Build.build_dir_kind with
       | Local    b -> Local.reach t ~from:(Local.append b from)
       | External _ -> external_of_in_source_tree t
     end
   | In_build_dir t, In_source_tree from -> begin
-      match Lazy.force build_dir_kind with
+      match Lazy.force Build.build_dir_kind with
       | Local    b -> Local.reach (Local.append b t) ~from
       | External b -> external_of_local t ~root:b
     end
   | In_source_tree t, External _ -> external_of_in_source_tree t
   | In_build_dir t, External _ ->
-    match Lazy.force build_dir_kind with
+    match Lazy.force Build.build_dir_kind with
     | Local    b -> external_of_in_source_tree (Local.append b t)
     | External b -> external_of_local t ~root:b
 
@@ -719,11 +869,14 @@ let is_descendant t ~of_ =
     Local.is_descendant t ~of_
   | _ -> false
 
-let append_local a b =
+let append_relative a b =
   match a with
-  | In_source_tree a -> in_source_tree (Local.append a b)
-  | In_build_dir a -> in_build_dir (Local.append a b)
-  | External a -> external_ (External.relative a (Local.to_string b))
+  | In_source_tree a -> in_source_tree (Relative.append a b)
+  | In_build_dir a -> in_build_dir (Relative.append a b)
+  | External a -> external_ (External.relative a (Relative.to_string b))
+
+let append_local = append_relative
+let append_source = append_relative
 
 let append a b =
   match b with
@@ -733,7 +886,7 @@ let append a b =
       [ "a", to_sexp a
       ; "b", to_sexp b
       ]
-  | In_source_tree b -> append_local a b
+  | In_source_tree b -> append_relative a b
 
 let basename t =
   match kind t with
@@ -741,15 +894,13 @@ let basename t =
   | External t -> External.basename t
 
 let parent = function
-  | External s ->
-    let parent = External.parent s in
-    if parent = s then
-      None
-    else
-      Some (external_ parent)
-  | In_source_tree p | In_build_dir p when Local.is_root p  -> None
-  | In_source_tree l -> Some (in_source_tree (Local.parent l))
-  | In_build_dir l -> Some (in_build_dir (Local.parent l))
+  | External s -> Option.map (External.parent s) ~f:external_
+  | In_source_tree l ->
+    Local.parent l
+    |> Option.map ~f:in_source_tree
+  | In_build_dir l ->
+    Local.parent l
+    |> Option.map ~f:in_build_dir
 
 let parent_exn t =
   match parent t with
@@ -772,27 +923,30 @@ let is_in_source_tree = function
   | In_build_dir _
   | External _ -> false
 
-let is_alias_stamp_file = function
-  | In_build_dir s -> String.is_prefix (Local.to_string s) ~prefix:".aliases/"
+let as_in_source_tree = function
+  | In_source_tree s -> Some s
+  | In_build_dir _
+  | External _ -> None
+
+let as_in_build_dir = function
+  | In_build_dir b -> Some b
   | In_source_tree _
-  | External _ -> false
+  | External _ -> None
+
+let as_in_build_dir_exn t = match t with
+  | External _ | In_source_tree _  ->
+    Exn.code_error
+      "[as_in_build_dir_exn] called on something not in build dir"
+      ["t", to_sexp t]
+  | In_build_dir p -> p
 
 let extract_build_context = function
   | In_source_tree _
   | External _ -> None
-  | In_build_dir p when Local.is_root p -> None
-  | In_build_dir t ->
-    let t = Local.to_string t in
-    begin match String.lsplit2 t ~on:'/' with
-    | None ->
-      Some (t, in_source_tree Local.root )
-    | Some (before, after) ->
-      Some
-        ( before
-        , after
-          |> Local.of_string
-          |> in_source_tree )
-    end
+  | In_build_dir p when Relative.is_root p -> None
+  | In_build_dir t -> Build.extract_build_context t
+
+let extract_build_dir_first_component = extract_build_context
 
 let extract_build_context_exn t =
   match extract_build_context t with
@@ -800,22 +954,12 @@ let extract_build_context_exn t =
   | None -> Exn.code_error "Path.extract_build_context_exn"
               ["t", to_sexp t]
 
-
 let extract_build_context_dir = function
   | In_source_tree _
   | External _ -> None
   | In_build_dir t ->
-    let t_str = Local.to_string t in
-    begin match String.lsplit2 t_str ~on:'/' with
-    | None -> Some (in_build_dir t, in_source_tree Local.root)
-    | Some (before, after) ->
-      Some
-        ( in_build_dir (Local.of_string before)
-        , after
-          |> Local.of_string
-          |> in_source_tree
-        )
-    end
+    Option.map (Build.extract_build_context_dir t)
+      ~f:(fun (base, rest) -> in_build_dir base, rest)
 
 let extract_build_context_dir_exn t =
   match extract_build_context_dir t with
@@ -834,16 +978,32 @@ let drop_build_context_exn t =
 let drop_optional_build_context t =
   match extract_build_context t with
   | None -> t
-  | Some (_, t) -> t
+  | Some (_, t) -> in_source_tree t
 
-let local_src   = Local.of_string "src"
-let local_build = Local.of_string "build"
-
-let sandbox_managed_paths ~sandbox_dir t =
+let drop_optional_build_context_src_exn t =
   match t with
-  | External _ -> t
-  | In_source_tree p -> append_local sandbox_dir (Local.append local_src   p)
-  | In_build_dir   p -> append_local sandbox_dir (Local.append local_build p)
+  | External _ ->
+    Exn.code_error "drop_optional_build_context_src_exn called on an external path" []
+  | In_build_dir _ ->
+    (match extract_build_context t with
+     | Some (_, s) -> s
+     | None ->
+       Exn.code_error
+         "drop_optional_build_context_src_exn called on a build directory itself" [])
+  | In_source_tree p -> p
+
+let local_src   = Relative.of_string "src"
+let local_build = Relative.of_string "build"
+
+let sandbox_managed_paths =
+  let append_local ~sandbox_dir local_x p =
+    in_build_dir (Build.append_relative sandbox_dir (Relative.append local_x p))
+  in
+  fun ~(sandbox_dir : Build.t) t ->
+    match t with
+    | External _ -> t
+    | In_source_tree p -> append_local ~sandbox_dir local_src p
+    | In_build_dir   p -> append_local ~sandbox_dir local_build p
 
 let split_first_component t =
   match kind t, is_root t with
@@ -875,7 +1035,26 @@ let explode_exn t =
 let exists t =
   try Sys.file_exists (to_string t)
   with Sys_error _ -> false
-let readdir_unsorted t = Sys.readdir (to_string t) |> Array.to_list
+let readdir_unsorted =
+  let rec loop dh acc =
+    match Unix.readdir dh with
+    | "."
+    | ".." -> loop dh acc
+    | s -> loop dh (s :: acc)
+    | exception End_of_file -> acc
+  in
+  fun t ->
+    try
+      let dh = Unix.opendir (to_string t) in
+      Exn.protect
+        ~f:(fun () ->
+          match loop dh [] with
+          | exception (Unix.Unix_error (e, _, _)) -> Error e
+          | s -> Result.Ok s)
+        ~finally:(fun () -> Unix.closedir dh)
+    with
+      Unix.Unix_error (e, _, _) -> Error e
+
 let is_directory t =
   try Sys.is_directory (to_string t)
   with Sys_error _ -> false
@@ -960,7 +1139,7 @@ let mkdir_p = function
   | In_source_tree s ->
     Local.mkdir_p s
   | In_build_dir k ->
-    Kind.mkdir_p (Kind.append_local (Lazy.force build_dir_kind) k)
+    Kind.mkdir_p (Kind.append_relative (Lazy.force Build.build_dir_kind) k)
 
 let compare x y =
   match x, y with
@@ -1010,13 +1189,13 @@ let pp_debug ppf = function
 module Set = struct
   include Set.Make(T)
   let to_sexp t = Sexp.Encoder.(list to_sexp) (to_list t)
-  let of_string_set ss ~f =
-    String.Set.to_list ss
-    |> List.map ~f
-    |> of_list
+  let of_listing ~dir ~filenames =
+    of_list (List.map filenames ~f:(fun f -> relative dir f))
 end
 
 let in_source s = in_source_tree (Local.of_string s)
+let source s = in_source_tree s
+let build s = in_build_dir s
 
 let is_suffix p ~suffix =
   match p with
@@ -1044,3 +1223,22 @@ let local_part = function
   | In_build_dir l -> l
 
 let stat t = Unix.stat (to_string t)
+
+include (Comparable.Operators(T) : Comparable.OPS with type t := t)
+
+module Source = struct
+  include Source0
+
+  let is_in_build_dir s =
+    is_in_build_dir (of_local s)
+
+  let to_local t = t
+end
+
+let set_of_source_paths set =
+  Source.Set.to_list set
+  |> List.map ~f:source
+  |> Set.of_list
+
+let set_of_build_paths_list =
+  List.fold_left ~init:Set.empty ~f:(fun acc e -> Set.add acc (build e))
